package authorizationheader

import (
	"crypto/x509"
	"fmt"
	"log"
	"sync"
	"time"

	"github.com/craigfurman/herottp"
)

type UserTokenAuthHeaderBuilder struct {
	uaaURL            string
	clientID          string
	clientSecret      string
	username          string
	password          string
	httpClient        HTTPClient
	cachedToken       string
	cachedTokenExpiry time.Time
	logger            *log.Logger
	tokenLock         *sync.Mutex
}

func NewUserTokenAuthHeaderBuilder(uaaURL, clientID, clientSecret, username, password string, disableSSLCertVerification bool, trustedCertPEM []byte, logger *log.Logger) *UserTokenAuthHeaderBuilder {
	rootCAs := x509.NewCertPool()
	rootCAs.AppendCertsFromPEM(trustedCertPEM)
	return &UserTokenAuthHeaderBuilder{
		uaaURL:       uaaURL,
		clientID:     clientID,
		clientSecret: clientSecret,
		username:     username,
		password:     password,
		httpClient: herottp.New(herottp.Config{
			DisableTLSCertificateVerification: disableSSLCertVerification,
			RootCAs: rootCAs,
			Timeout: 30 * time.Second,
		}),
		logger:    logger,
		tokenLock: new(sync.Mutex),
	}
}

func (hb *UserTokenAuthHeaderBuilder) Build() (string, error) {
	hb.tokenLock.Lock()
	defer hb.tokenLock.Unlock()

	var err error
	hb.cachedToken, hb.cachedTokenExpiry, err = getValidToken(hb.cachedToken, hb.cachedTokenExpiry, hb.logger, hb.obtainToken)
	if err != nil {
		return "", err
	}

	return bearerToken(hb.cachedToken), nil
}

func (hb *UserTokenAuthHeaderBuilder) obtainToken() (string, int, error) {
	requestBody := fmt.Sprintf("grant_type=password&username=%s&password=%s", hb.username, hb.password)

	request, err := buildTokenRequest(requestBody, hb.uaaURL, hb.clientID, hb.clientSecret)
	if err != nil {
		return "", 0, err
	}

	return doObtainTokenRequest(hb.httpClient, hb.logger, request)
}
