package broker

import (
	"encoding/json"
	"errors"
	"fmt"
	"log"
	"sync"

	"github.com/pivotal-cf/brokerapi"
	"github.com/pivotal-cf/on-demand-service-broker-sdk/bosh"
	"github.com/pivotal-cf/on-demand-service-broker-sdk/serviceadapter"
	"github.com/pivotal-cf/on-demand-service-broker/boshclient"
)

//go:generate counterfeiter -o fakes/fake_bosh_client.go . BoshClient
type BoshClient interface {
	DeployRaw(manifest []byte) (int, error)
	GetTask(taskID int) (boshclient.BoshTask, error)
	GetTasks(deploymentName string) ([]boshclient.BoshTask, error)
	VMs(instanceID string) (bosh.BoshVMs, error)
	DeleteDeployment(name string) (int, error)
	GetManifestRaw(name string) ([]byte, error)
}

type OperationType string

const (
	OperationTypeCreate  = OperationType("Create")
	OperationTypeUpdate  = OperationType("Update")
	OperationTypeUpgrade = OperationType("Upgrade")
	OperationTypeDelete  = OperationType("Delete")
)

type OperationData struct {
	BoshTaskID    int
	OperationType OperationType
}

//go:generate counterfeiter -o fakes/fake_deployer.go . Deployer
type Deployer interface {
	Deploy(instanceID, planID, deploymentName string, operationType OperationType, arbitraryParams map[string]interface{}, oldManifest []byte, previousPlanID *string) (int, []byte, error)
}

//go:generate counterfeiter -o fakes/fake_service_adapter_client.go . ServiceAdapterClient
type ServiceAdapterClient interface {
	GenerateManifest(serviceReleases serviceadapter.ServiceDeployment, plan serviceadapter.Plan, requestParams map[string]interface{}, previousManifest []byte, previousPlan *serviceadapter.Plan) ([]byte, error)
	CreateBinding(bindingID string, deploymentTopology bosh.BoshVMs, manifest []byte, requestParams map[string]interface{}) (serviceadapter.Binding, error)
	DeleteBinding(bindingID string, deploymentTopology bosh.BoshVMs, manifest []byte, requestParams map[string]interface{}) error
	GenerateDashboardUrl(instanceID string, plan serviceadapter.Plan, manifest []byte) (string, error)
}

type InstanceState struct {
	PlanID              string
	OperationInProgress bool
}

//go:generate counterfeiter -o fakes/fake_cloud_foundry_client.go . CloudFoundryClient
type CloudFoundryClient interface {
	CountInstancesOfServiceOffering(serviceOfferingID string) (instanceCountByPlanID map[string]int, err error)
	CountInstancesOfPlan(serviceOfferingID, planID string) (int, error)
	GetInstanceState(serviceInstanceGUID string) (InstanceState, error)
	GetInstancesOfServiceOffering(serviceOfferingID string) ([]string, error)
}

type NotFoundError struct {
	error
}

func NewNotFoundError() NotFoundError {
	return NotFoundError{error: errors.New("resource was not found")}
}

func IsNotFound(err error) bool {
	switch err.(type) {
	case NotFoundError:
		return true
	default:
		return false
	}
}

type ServiceAdapterCommandNotImplementedError struct {
	error
}

func NewServiceAdapterCommandNotImplementedError(e error) ServiceAdapterCommandNotImplementedError {
	return ServiceAdapterCommandNotImplementedError{e}
}

type Broker struct {
	boshClient     BoshClient
	cfClient       CloudFoundryClient
	serviceAdapter ServiceAdapterClient
	deployer       Deployer
	deploymentLock *sync.Mutex

	serviceOffering ServiceCatalog

	logger *log.Logger
}

func New(boshClient BoshClient, cfClient CloudFoundryClient, serviceAdapter ServiceAdapterClient,
	deployer Deployer, serviceOffering ServiceCatalog, logger *log.Logger) (*Broker, error) {

	plansToInstanceCounts, err := cfClient.CountInstancesOfServiceOffering(serviceOffering.ID)
	if err != nil {
		return nil, err
	}

	for existingPlanID, existingPlanCount := range plansToInstanceCounts {
		if existingPlanCount > 0 && !serviceOffering.Plans.Contains(existingPlanID) {
			return nil, fmt.Errorf("You cannot change the plan_id of a plan that has existing service instances")
		}
	}

	return &Broker{
		boshClient:     boshClient,
		cfClient:       cfClient,
		serviceAdapter: serviceAdapter,
		deployer:       deployer,
		deploymentLock: &sync.Mutex{},

		serviceOffering: serviceOffering,

		logger: logger,
	}, nil
}

func convertProvisionDetailsToMap(details brokerapi.ProvisionDetails) (map[string]interface{}, error) {
	requestParams, err := convertToMap(details)
	if err != nil {
		return nil, err
	}

	if len(details.RawParameters) > 0 {
		arbitraryParams := map[string]interface{}{}
		if err := json.Unmarshal(details.RawParameters, &arbitraryParams); err != nil {
			return nil, brokerapi.ErrRawParamsInvalid
		}
		requestParams["parameters"] = arbitraryParams
	}

	return requestParams, nil
}

func convertToMap(object interface{}) (map[string]interface{}, error) {
	genericMap := map[string]interface{}{}
	data, err := json.Marshal(object)
	if err != nil {
		return nil, err
	}
	err = json.Unmarshal(data, &genericMap)

	if err != nil {
		return nil, err
	}
	return genericMap, nil
}

type DisplayableError struct {
	WrappedError error
}

func (e DisplayableError) Error() string {
	return e.WrappedError.Error()
}

type AdapterCommandError struct {
	messageForCLIUser  string
	messageForOperator string
}

func (e AdapterCommandError) Error() string {
	return e.messageForOperator
}

func (e AdapterCommandError) MessageForCLIUser() string {
	return e.messageForCLIUser
}

func NewAdapterCommandError(messageForCLIUser, messageForOperator string) AdapterCommandError {
	return AdapterCommandError{messageForCLIUser: messageForCLIUser, messageForOperator: messageForOperator}
}

func convertToDisplayableError(err error, logger *log.Logger) error {
	logger.Printf("error: %s\n", err)

	switch err := err.(type) {
	case nil:
		return nil
	case DisplayableError:
		return err.WrappedError
	case AdapterCommandError:
		return errors.New(err.MessageForCLIUser())
	default:
		return errors.New(genericErrorMessage)
	}
}

const genericErrorMessage = "An internal error has occurred"
