package broker

import (
	"fmt"
	"log"

	"github.com/pivotal-cf/on-demand-service-broker-sdk/serviceadapter"
	"github.com/pivotal-cf/on-demand-service-broker/boshclient"
	"github.com/pivotal-cf/on-demand-service-broker/mgmtapi"
)

type deployer struct {
	boshClient      BoshClient
	serviceAdapter  ServiceAdapterClient
	serviceOffering ServiceCatalog
	serviceStemcell serviceadapter.Stemcell
	serviceReleases serviceadapter.ServiceReleases
	logger          *log.Logger
}

func NewDeployer(
	boshClient BoshClient,
	serviceAdapter ServiceAdapterClient,
	serviceOffering ServiceCatalog,
	serviceStemcell serviceadapter.Stemcell,
	serviceReleases serviceadapter.ServiceReleases,
	logger *log.Logger,
) Deployer {
	return deployer{
		boshClient:      boshClient,
		serviceAdapter:  serviceAdapter,
		serviceOffering: serviceOffering,
		serviceStemcell: serviceStemcell,
		serviceReleases: serviceReleases,
		logger:          logger,
	}
}

func (d deployer) Deploy(instanceID, planID, deploymentName string, operationType OperationType, requestParams map[string]interface{}, oldManifest []byte, previousPlanID *string) (int, []byte, error) {
	tasks, err := d.boshClient.GetTasks(deploymentName)
	if err != nil {
		d.logger.Printf("error getting last operation for deployment %s: %s\n", deploymentName, err)
		return 0, nil, err
	}

	if deploymentInProgress(tasks) {
		incompleteTasks := selectIncompleteTasks(tasks)
		d.logger.Printf("deployment %s is still in progress: %v\n", deploymentName, incompleteTasks)
		return 0, nil, mgmtapi.NewOperationInProgressError(fmt.Errorf("deployment %s is still in progress", deploymentName))
	}

	manifest, err := d.generateManifestForDeployment(deploymentName, instanceID, planID, requestParams, oldManifest, previousPlanID)
	if err != nil {
		d.logger.Printf("error generating manifest: %s\n", err)
		return 0, nil, err
	}

	taskID, err := d.boshClient.DeployRaw(manifest)
	if err != nil {
		return 0, nil, fmt.Errorf("error deploying instance: %s\n", err)
	}
	d.logger.Printf("Bosh task ID for %s instance %s is %d\n", operationType, instanceID, taskID)

	return taskID, manifest, nil
}

func serviceAdapterPlan(plan Plan) serviceadapter.Plan {
	abridgedPlan := serviceadapter.Plan{
		Properties:     plan.Properties,
		InstanceGroups: plan.InstanceGroups,
		Update:         plan.Update,
	}
	return abridgedPlan
}

func (d deployer) findPlans(planID string, previousPlanID *string) (serviceadapter.Plan, *serviceadapter.Plan, error) {
	plan := d.serviceOffering.FindPlanById(planID)
	if plan == nil {
		return serviceadapter.Plan{}, nil, DisplayableError{fmt.Errorf("plan %s does not exist", planID)}
	}
	var previousAdapterPlan *serviceadapter.Plan
	if previousPlanID != nil {
		previousPlan := d.serviceOffering.FindPlanById(*previousPlanID)
		if previousPlan == nil {
			return serviceadapter.Plan{}, nil, DisplayableError{fmt.Errorf("previous plan %s does not exist", *previousPlanID)}
		}
		convertedPlan := serviceAdapterPlan(*previousPlan)
		previousAdapterPlan = &convertedPlan
	}

	return serviceAdapterPlan(*plan), previousAdapterPlan, nil
}

func deploymentInProgress(tasks []boshclient.BoshTask) bool {
	for _, task := range tasks {
		if task.State == boshclient.BoshTaskQueued || task.State == boshclient.BoshTaskProcessing {
			return true
		}
	}

	return false
}

func selectIncompleteTasks(tasks []boshclient.BoshTask) []boshclient.BoshTask {
	var incompleteTasks []boshclient.BoshTask
	for _, task := range tasks {
		if task.State == boshclient.BoshTaskQueued || task.State == boshclient.BoshTaskProcessing {
			incompleteTasks = append(incompleteTasks, task)
		}
	}
	return incompleteTasks
}

func (d deployer) generateManifestForDeployment(deploymentName string, instanceID string, planID string, requestParams map[string]interface{}, oldManifest []byte, previousPlanID *string) ([]byte, error) {
	plan, previousPlan, err := d.findPlans(planID, previousPlanID)
	if err != nil {
		return nil, err
	}

	serviceDeployment := serviceadapter.ServiceDeployment{DeploymentName: deploymentName, Releases: d.serviceReleases, Stemcell: d.serviceStemcell}

	d.logger.Printf("service adapter will generate manifest for instance %s\n", instanceID)
	return d.serviceAdapter.GenerateManifest(serviceDeployment, plan, requestParams, oldManifest, previousPlan)
}
