package broker_test

import (
	"bytes"
	"errors"
	"fmt"
	"io"
	"log"

	. "github.com/onsi/ginkgo"
	. "github.com/onsi/gomega"
	"github.com/pivotal-cf/on-demand-service-broker-sdk/serviceadapter"
	"github.com/pivotal-cf/on-demand-service-broker/boshclient"
	"github.com/pivotal-cf/on-demand-service-broker/broker"
	"github.com/pivotal-cf/on-demand-service-broker/mgmtapi"
)

var _ = Describe("Deployer", func() {

	var (
		deployer          broker.Deployer
		planID            string
		previousPlanID    *string
		arbitraryParams   map[string]interface{}
		oldManifest       []byte
		deployedManifest  []byte
		deployError       error
		operationType     broker.OperationType
		inMemoryLogBuffer *bytes.Buffer
		boshTaskID        = 42
		returnedTaskID    int
		instanceID        = "some-instance-id"

		manifest          = []byte("a-manifest")
		deploymentName    = "a-fancy-deployment"
		serviceDeployment serviceadapter.ServiceDeployment
	)

	BeforeEach(func() {
		serviceDeployment = serviceadapter.ServiceDeployment{
			Releases: serviceadapter.ServiceReleases{{
				Name:    "name",
				Version: "vers",
				Jobs:    []string{"a", "b"},
			}},
			Stemcell: serviceadapter.Stemcell{
				OS:      "RISC OS",
				Version: "2",
			},
		}
		inMemoryLogBuffer = bytes.NewBuffer([]byte{})
		combinedWriter := io.MultiWriter(inMemoryLogBuffer, GinkgoWriter)
		logger := log.New(combinedWriter, "[unit-tests] ", log.LstdFlags)
		deployer = broker.NewDeployer(boshClient, serviceAdapter, serviceCatalog, serviceDeployment.Stemcell, serviceDeployment.Releases, logger)
		planID = existingPlanID
		operationType = broker.OperationTypeCreate
		previousPlanID = stringPointer(secondPlanID)
	})

	JustBeforeEach(func() {
		returnedTaskID, deployedManifest, deployError = deployer.Deploy(instanceID, planID, deploymentName, operationType, arbitraryParams, oldManifest, previousPlanID)
	})

	Context("deployment without previous manifest", func() {
		BeforeEach(func() {
			oldManifest = nil
			previousPlanID = nil
		})
		Context("when bosh deploys the release successfully", func() {
			BeforeEach(func() {
				boshClient.DeployRawReturns(42, nil)
				serviceAdapter.GenerateManifestReturns(manifest, nil)
				By("not having any previous tasks")
				boshClient.GetTasksReturns([]boshclient.BoshTask{}, nil)
			})

			It("checks tasks for the deployment", func() {
				Expect(boshClient.GetTasksCallCount()).To(Equal(1))
				Expect(boshClient.GetTasksArgsForCall(0)).To(Equal(deploymentName))
			})

			It("generates a manifest", func() {
				Expect(serviceAdapter.GenerateManifestCallCount()).To(Equal(1))
				actualServiceDeployment, actualPlan, actualArbParams, actualPreviousManifest, acutalPreviousPlan := serviceAdapter.GenerateManifestArgsForCall(0)
				Expect(actualArbParams).To(Equal(arbitraryParams))
				Expect(actualPlan).To(Equal(serviceadapter.Plan{InstanceGroups: existingPlan.InstanceGroups, Properties: existingPlan.Properties, Update: existingPlan.Update}))
				Expect(actualServiceDeployment).To(Equal(serviceadapter.ServiceDeployment{DeploymentName: deploymentName, Releases: serviceDeployment.Releases, Stemcell: serviceDeployment.Stemcell}))
				Expect(actualPreviousManifest).To(BeNil())
				Expect(acutalPreviousPlan).To(BeNil())
			})

			It("returns the bosh task ID", func() {
				Expect(returnedTaskID).To(Equal(boshTaskID))
			})

			It("Creates a bosh deployment using generated manifest", func() {
				Expect(boshClient.DeployRawCallCount()).To(Equal(1))
				deployedManifest := boshClient.DeployRawArgsForCall(0)
				Expect(deployedManifest).To(Equal(manifest))
			})

			It("return the newly generated manifest", func() {
				Expect(deployedManifest).To(Equal(manifest))
			})
		})
	})
	Context("logging", func() {
		BeforeEach(func() {
			boshClient.DeployRawReturns(42, nil)
			serviceAdapter.GenerateManifestReturns(manifest, nil)
			boshClient.GetTasksReturns([]boshclient.BoshTask{{State: boshclient.BoshTaskDone}}, nil)
		})
		Context("for creating", func() {
			BeforeEach(func() {
				operationType = broker.OperationTypeCreate
			})
			It("logs the bosh task ID returned by the director", func() {
				Expect(inMemoryLogBuffer.String()).To(ContainSubstring(fmt.Sprintf("Bosh task ID for Create instance %s is %d", instanceID, boshTaskID)))
			})
		})
		Context("for updating", func() {
			BeforeEach(func() {
				operationType = broker.OperationTypeUpdate
			})
			It("logs the bosh task ID returned by the director", func() {
				Expect(inMemoryLogBuffer.String()).To(ContainSubstring(fmt.Sprintf("Bosh task ID for Update instance %s is %d", instanceID, boshTaskID)))
			})
		})
		Context("for upgrading", func() {
			BeforeEach(func() {
				operationType = broker.OperationTypeUpgrade
			})
			It("logs the bosh task ID returned by the director", func() {
				Expect(inMemoryLogBuffer.String()).To(ContainSubstring(fmt.Sprintf("Bosh task ID for Upgrade instance %s is %d", instanceID, boshTaskID)))
			})
		})
	})

	Context("deployment with previous manifest", func() {
		BeforeEach(func() {
			oldManifest = []byte("old-manifest-fetched-from-bosh")
			boshClient.DeployRawReturns(42, nil)
			serviceAdapter.GenerateManifestReturns(manifest, nil)
			boshClient.GetTasksReturns([]boshclient.BoshTask{{State: boshclient.BoshTaskDone}}, nil)
			arbitraryParams = map[string]interface{}{"foo": "bar"}
		})

		It("checks tasks for the deployment", func() {
			Expect(boshClient.GetTasksCallCount()).To(Equal(1))
			Expect(boshClient.GetTasksArgsForCall(0)).To(Equal(deploymentName))
		})

		It("Creates a bosh deployment from the generated manifest", func() {
			Expect(boshClient.DeployRawCallCount()).To(Equal(1))
			deployedManifest := boshClient.DeployRawArgsForCall(0)
			Expect(deployedManifest).To(Equal(manifest))
		})

		It("generates manifest with correct inputs", func() {
			Expect(serviceAdapter.GenerateManifestCallCount()).To(Equal(1))
			actualServiceDeployment, planPassedToGenerate, arbParamsPassedToGenerate, previousManifestPassedToGenerate, previousPlanProperties := serviceAdapter.GenerateManifestArgsForCall(0)
			Expect(planPassedToGenerate).To(Equal(serviceadapter.Plan{InstanceGroups: existingPlan.InstanceGroups, Properties: existingPlan.Properties, Update: existingPlan.Update}))
			Expect(arbParamsPassedToGenerate).To(Equal(arbitraryParams))
			Expect(previousManifestPassedToGenerate).To(Equal(oldManifest))
			Expect(actualServiceDeployment).To(Equal(serviceadapter.ServiceDeployment{DeploymentName: deploymentName, Releases: serviceDeployment.Releases, Stemcell: serviceDeployment.Stemcell}))
			Expect(previousPlanProperties).To(Equal(&serviceadapter.Plan{InstanceGroups: secondPlan.InstanceGroups, Properties: secondPlan.Properties}))
		})

		It("returns the bosh task ID", func() {
			Expect(returnedTaskID).To(Equal(boshTaskID))
		})
	})
	Context("plan isnt found", func() {
		BeforeEach(func() {
			planID = "non-existant-plan"
			boshClient.GetTasksReturns([]boshclient.BoshTask{{State: boshclient.BoshTaskDone}}, nil)
			arbitraryParams = map[string]interface{}{"foo": "bar"}
		})
		It("checks tasks for the deployment", func() {
			Expect(boshClient.GetTasksCallCount()).To(Equal(1))
			Expect(boshClient.GetTasksArgsForCall(0)).To(Equal(deploymentName))
		})

		It("does not deploy", func() {
			Expect(boshClient.DeployRawCallCount()).To(BeZero())
		})

		It("returns an error", func() {
			Expect(deployError).To(HaveOccurred())
			Expect(deployError).To(MatchError(ContainSubstring("plan non-existant-plan does not exist")))
		})
	})

	Context("previous plan isnt found", func() {
		BeforeEach(func() {
			previousPlanID = stringPointer("non-existant-plan")
			boshClient.GetTasksReturns([]boshclient.BoshTask{{State: boshclient.BoshTaskDone}}, nil)
			arbitraryParams = map[string]interface{}{"foo": "bar"}
		})
		It("checks tasks for the deployment", func() {
			Expect(boshClient.GetTasksCallCount()).To(Equal(1))
			Expect(boshClient.GetTasksArgsForCall(0)).To(Equal(deploymentName))
		})

		It("does not deploy", func() {
			Expect(boshClient.DeployRawCallCount()).To(BeZero())
		})

		It("returns an error", func() {
			Expect(deployError).To(HaveOccurred())
			Expect(deployError).To(MatchError(ContainSubstring("previous plan non-existant-plan does not exist")))
		})
	})

	Context("when the last bosh task for deployment is queued", func() {
		var previousDoneBoshTaskID = 41
		var previousErrorBoshTaskID = 40

		BeforeEach(func() {
			boshClient.GetTasksReturns([]boshclient.BoshTask{
				{State: boshclient.BoshTaskQueued, ID: boshTaskID},
				{State: boshclient.BoshTaskDone, ID: previousDoneBoshTaskID},
				{State: boshclient.BoshTaskError, ID: previousErrorBoshTaskID},
			}, nil)
		})

		It("returns an error", func() {
			Expect(deployError).To(MatchError(ContainSubstring(fmt.Sprintf("deployment %s is still in progress", deploymentName))))
			Expect(deployError).To(BeAssignableToTypeOf(mgmtapi.OperationInProgressError{}))
		})

		It("logs the queued task", func() {
			Expect(inMemoryLogBuffer.String()).To(ContainSubstring("deployment %s is still in progress:", deploymentName))
			Expect(inMemoryLogBuffer.String()).To(ContainSubstring("{%d", boshTaskID))
		})

		It("does not log the previous completed tasks for the deployment", func() {
			Expect(inMemoryLogBuffer.String()).NotTo(ContainSubstring("done"))
			Expect(inMemoryLogBuffer.String()).NotTo(ContainSubstring("{%d", previousDoneBoshTaskID))
			Expect(inMemoryLogBuffer.String()).NotTo(ContainSubstring("error"))
			Expect(inMemoryLogBuffer.String()).NotTo(ContainSubstring("{%d", previousErrorBoshTaskID))
		})
	})

	Context("when the last bosh task for deployment is processing", func() {
		var previousDoneBoshTaskID = 41
		var previousErrorBoshTaskID = 40

		BeforeEach(func() {
			boshClient.GetTasksReturns([]boshclient.BoshTask{
				{State: boshclient.BoshTaskProcessing, ID: boshTaskID},
				{State: boshclient.BoshTaskDone, ID: previousDoneBoshTaskID},
				{State: boshclient.BoshTaskError, ID: previousErrorBoshTaskID},
			}, nil)
		})

		It("returns an error", func() {
			Expect(deployError).To(MatchError(ContainSubstring(fmt.Sprintf("deployment %s is still in progress", deploymentName))))
			Expect(deployError).To(BeAssignableToTypeOf(mgmtapi.OperationInProgressError{}))
		})

		It("logs the queued task", func() {
			Expect(inMemoryLogBuffer.String()).To(ContainSubstring("deployment %s is still in progress:", deploymentName))
			Expect(inMemoryLogBuffer.String()).To(ContainSubstring("{%d", boshTaskID))
		})

		It("does not log the previous tasks for the deployment", func() {
			Expect(inMemoryLogBuffer.String()).NotTo(ContainSubstring("done"))
			Expect(inMemoryLogBuffer.String()).NotTo(ContainSubstring("{%d", previousDoneBoshTaskID))
			Expect(inMemoryLogBuffer.String()).NotTo(ContainSubstring("error"))
			Expect(inMemoryLogBuffer.String()).NotTo(ContainSubstring("{%d", previousErrorBoshTaskID))
		})
	})

	Context("when the last bosh task for deployment fails to fetch", func() {
		BeforeEach(func() {
			boshClient.GetTasksReturns(nil, fmt.Errorf("connection error"))
		})
		It("wraps the error", func() {
			Expect(deployError).To(MatchError(ContainSubstring("connection error")))
		})
	})

	Context("when the manifest generator fails to generate the manifest", func() {
		BeforeEach(func() {
			boshClient.GetTasksReturns([]boshclient.BoshTask{{State: boshclient.BoshTaskDone}}, nil)
			serviceAdapter.GenerateManifestReturns(nil, errors.New("manifest fail"))
		})

		It("wraps the error", func() {
			Expect(deployError).To(MatchError(ContainSubstring("manifest fail")))
		})
	})

	Context("when bosh fails to deploy the release", func() {
		BeforeEach(func() {
			boshClient.GetTasksReturns([]boshclient.BoshTask{{State: boshclient.BoshTaskDone}}, nil)
			boshClient.DeployRawReturns(0, errors.New("oops"))
		})

		It("wraps the error", func() {
			Expect(deployError).To(MatchError(ContainSubstring("oops")))
		})
	})
})

func stringPointer(s string) *string {
	return &s
}
