package broker_test

import (
	"encoding/json"
	"errors"
	"fmt"

	. "github.com/onsi/ginkgo"
	. "github.com/onsi/gomega"
	"github.com/pivotal-cf/brokerapi"
	"github.com/pivotal-cf/on-demand-service-broker/boshclient"
	"github.com/pivotal-cf/on-demand-service-broker/broker"
)

var _ = Describe("deprovisioning instances", func() {
	var (
		instanceID = "an-instance-to-be-deprovisioned"

		asyncAllowed    bool
		deprovisionSpec brokerapi.DeprovisionServiceSpec
		deprovisionErr  error

		deleteTaskID = 88
	)

	BeforeEach(func() {
		asyncAllowed = true
	})

	BeforeEach(func() {
		boshClient.DeleteDeploymentReturns(deleteTaskID, nil)
	})

	JustBeforeEach(func() {
		deprovisionSpec, deprovisionErr = b.Deprovision(instanceID, brokerapi.DeprovisionDetails{}, asyncAllowed)
	})

	It("returns that is deprovisioning asynchronously", func() {
		Expect(deprovisionSpec.IsAsync).To(BeTrue())
	})

	It("returns the bosh task ID and operation type in the operation data", func() {
		var operationData broker.OperationData
		Expect(json.Unmarshal([]byte(deprovisionSpec.OperationData), &operationData)).To(Succeed())
		Expect(operationData).To(Equal(broker.OperationData{BoshTaskID: deleteTaskID, OperationType: broker.OperationTypeDelete}))
	})

	It("deletes a bosh deployment whose name is based on the instance ID", func() {
		Expect(boshClient.DeleteDeploymentCallCount()).To(Equal(1))
		Expect(boshClient.DeleteDeploymentArgsForCall(0)).To(Equal(deploymentName(instanceID)))
	})

	It("logs that it will delete the deployment", func() {
		Expect(logBuffer).To(ContainSubstring(fmt.Sprintf("deleting deployment for instance %s", instanceID)))
	})

	It("logs the task id to delete the deployment", func() {
		Expect(logBuffer.String()).To(ContainSubstring(fmt.Sprintf("Bosh task id for Delete instance %s was %d", instanceID, deleteTaskID)))
	})

	Context("when the async allowed flag is false", func() {
		BeforeEach(func() {
			asyncAllowed = false
		})

		It("returns a error", func() {
			Expect(deprovisionErr).To(Equal(brokerapi.ErrAsyncRequired))
		})
	})

	Context("when the bosh client cannot delete the deployment", func() {
		BeforeEach(func() {
			boshClient.DeleteDeploymentReturns(0, errors.New("er ma gerd!"))
		})

		It("returns a generic error message", func() {
			Expect(deprovisionErr).To(MatchError("An internal error has occurred"))
		})

		It("logs the error", func() {
			Expect(logBuffer.String()).To(ContainSubstring("er ma gerd!"))
		})
	})

	Context("when the bosh client cannot find the deployment", func() {
		BeforeEach(func() {
			boshClient.DeleteDeploymentReturns(0, boshclient.DeploymentNotFoundError{})
		})

		It("returns an error", func() {
			Expect(deprovisionErr).To(Equal(brokerapi.ErrInstanceDoesNotExist))
		})

		It("logs the error", func() {
			Expect(logBuffer.String()).To(ContainSubstring(fmt.Sprintf("error deprovisioning: instance %s, not found", instanceID)))
		})
	})
})
