package broker_test

import (
	"encoding/json"
	"errors"
	"fmt"

	"github.com/pivotal-cf/brokerapi"
	"github.com/pivotal-cf/on-demand-service-broker/boshclient"
	"github.com/pivotal-cf/on-demand-service-broker/broker"

	. "github.com/onsi/ginkgo"
	. "github.com/onsi/gomega"
)

var _ = Describe("LastOperation", func() {
	Context("failures", func() {
		var (
			instanceID    = "a-useful-instance"
			operationData string

			lastOpErr error
		)

		BeforeEach(func() {
			operationData = "{}"
		})

		JustBeforeEach(func() {
			_, lastOpErr = b.LastOperation(instanceID, operationData)
		})

		Context("when task cannot be retrieved from BOSH", func() {
			BeforeEach(func() {
				operationData = `{"BoshTaskID": 42}`
				boshClient.GetTaskReturns(boshclient.BoshTask{}, errors.New("something went wrong!"))
			})

			It("logs the error", func() {
				Expect(logBuffer.String()).To(ContainSubstring("error retrieving tasks from bosh, for deployment 'service-instance_a-useful-instance': something went wrong!"))
			})

			It("returns a generic error", func() {
				Expect(lastOpErr).To(MatchError("An internal error has occurred"))
			})
		})

		Context("when there is no task ID in the operation data", func() {
			BeforeEach(func() {
				operationData = "{}"
			})

			It("logs the error", func() {
				Expect(logBuffer.String()).To(ContainSubstring("no task ID found in operation data"))
			})

			It("returns a generic error", func() {
				Expect(lastOpErr).To(MatchError("An internal error has occurred"))
			})
		})

		Context("when operation data cannot be parsed", func() {
			BeforeEach(func() {
				operationData = "I'm not JSON"
			})

			It("logs the error", func() {
				Expect(logBuffer.String()).To(ContainSubstring("operation data cannot be parsed"))
			})

			It("returns a generic error", func() {
				Expect(lastOpErr).To(MatchError("An internal error has occurred"))
			})
		})
	})

	Context("when the task can be retrieved", func() {
		var (
			instanceID    = "not-relevent"
			operationData []byte
			taskID        = 199
		)

		type testCase struct {
			ActualBoshTask      boshclient.BoshTask
			ActualOperationType broker.OperationType
			LogContains         string

			ExpectedLastOperation brokerapi.LastOperation
			TaskIDRemoved         bool
		}

		testLastOperation := func(testCase testCase) func() {
			return func() {
				var (
					actualLastOperation      brokerapi.LastOperation
					actualLastOperationError error
				)

				JustBeforeEach(func() {
					var err error
					operationData, err = json.Marshal(broker.OperationData{OperationType: testCase.ActualOperationType, BoshTaskID: taskID})
					Expect(err).NotTo(HaveOccurred())

					boshClient.GetTaskReturns(testCase.ActualBoshTask, nil)
					actualLastOperation, actualLastOperationError = b.LastOperation(instanceID, string(operationData))
				})

				It("retrieves the task by ID", func() {
					Expect(boshClient.GetTaskCallCount()).To(Equal(1))
					Expect(boshClient.GetTaskArgsForCall(0)).To(Equal(taskID))
				})

				It("does not error", func() {
					Expect(actualLastOperationError).NotTo(HaveOccurred())
				})

				It("returns the state", func() {
					Expect(actualLastOperation.State).To(Equal(testCase.ExpectedLastOperation.State))
				})

				It("returns the description", func() {
					Expect(actualLastOperation.Description).To(Equal(testCase.ExpectedLastOperation.Description))
				})

				It("logs a message", func() {
					Expect(logBuffer.String()).To(ContainSubstring(testCase.LogContains))
				})

				if testCase.ActualOperationType == broker.OperationTypeCreate {
					if testCase.ActualBoshTask.State == boshclient.BoshTaskDone {
						It("logs that the deployment has been created", func() {
							Expect(logBuffer).To(ContainSubstring(fmt.Sprintf("BOSH task ID %d status: success creating deployment for instance %s: Description: %s Result: %s\n", taskID, instanceID, testCase.ActualBoshTask.Description, testCase.ActualBoshTask.Result)))
						})
					} else if testCase.ActualBoshTask.State == boshclient.BoshTaskError {
						It("logs that the deployment has been failed to create", func() {
							Expect(logBuffer).To(ContainSubstring(fmt.Sprintf("BOSH task ID %d status: failure creating deployment for instance %s: Description: %s Result: %s\n", taskID, instanceID, testCase.ActualBoshTask.Description, testCase.ActualBoshTask.Result)))
						})
					}
				} else if testCase.ActualOperationType == broker.OperationTypeDelete {
					if testCase.ActualBoshTask.State == boshclient.BoshTaskDone {
						It("logs that the deployment has been deleted", func() {
							Expect(logBuffer).To(ContainSubstring(fmt.Sprintf("BOSH task ID %d status: success deleting deployment for instance %s: Description: %s Result: %s\n", taskID, instanceID, testCase.ActualBoshTask.Description, testCase.ActualBoshTask.Result)))
						})
					} else if testCase.ActualBoshTask.State == boshclient.BoshTaskError {
						It("logs that the deployment has been failed to delete", func() {
							Expect(logBuffer).To(ContainSubstring(fmt.Sprintf("BOSH task ID %d status: failure deleting deployment for instance %s: Description: %s Result: %s\n", taskID, instanceID, testCase.ActualBoshTask.Description, testCase.ActualBoshTask.Result)))
						})
					}
				} else if testCase.ActualOperationType == broker.OperationTypeUpdate {
					if testCase.ActualBoshTask.State == boshclient.BoshTaskDone {
						It("logs that the deployment has been updated", func() {
							Expect(logBuffer).To(ContainSubstring(fmt.Sprintf("BOSH task ID %d status: success updating deployment for instance %s: Description: %s Result: %s\n", taskID, instanceID, testCase.ActualBoshTask.Description, testCase.ActualBoshTask.Result)))
						})
					} else if testCase.ActualBoshTask.State == boshclient.BoshTaskError {
						It("logs that the deployment has been failed to update", func() {
							Expect(logBuffer).To(ContainSubstring(fmt.Sprintf("BOSH task ID %d status: failure updating deployment for instance %s: Description: %s Result: %s\n", taskID, instanceID, testCase.ActualBoshTask.Description, testCase.ActualBoshTask.Result)))
						})
					}
				} else if testCase.ActualOperationType == broker.OperationTypeUpgrade {
					if testCase.ActualBoshTask.State == boshclient.BoshTaskDone {
						It("logs that the deployment has been upgraded", func() {
							Expect(logBuffer).To(ContainSubstring(fmt.Sprintf("BOSH task ID %d status: success upgrading deployment for instance %s: Description: %s Result: %s\n", taskID, instanceID, testCase.ActualBoshTask.Description, testCase.ActualBoshTask.Result)))
						})
					} else if testCase.ActualBoshTask.State == boshclient.BoshTaskError {
						It("logs that the deployment has been failed to upgrade", func() {
							Expect(logBuffer).To(ContainSubstring(fmt.Sprintf("BOSH task ID %d status: failure upgrading deployment for instance %s: Description: %s Result: %s\n", taskID, instanceID, testCase.ActualBoshTask.Description, testCase.ActualBoshTask.Result)))
						})
					}
				}
			}
		}
		Describe("while creating ", func() {
			Describe("last operation is Processing",
				testLastOperation(testCase{
					ActualBoshTask:      boshclient.BoshTask{State: boshclient.BoshTaskProcessing, Description: "it's a task", ID: taskID},
					ActualOperationType: broker.OperationTypeCreate,

					ExpectedLastOperation: brokerapi.LastOperation{
						State:       brokerapi.InProgress,
						Description: broker.CreateInProgress,
					},
				}))
			Describe("last operation is Queued",
				testLastOperation(testCase{
					ActualBoshTask:      boshclient.BoshTask{State: boshclient.BoshTaskQueued, Description: "it's a task", ID: taskID},
					ActualOperationType: broker.OperationTypeCreate,

					ExpectedLastOperation: brokerapi.LastOperation{
						State:       brokerapi.InProgress,
						Description: broker.CreateInProgress,
					},
				}))

			Describe("last operation is Error",
				testLastOperation(testCase{
					ActualBoshTask:      boshclient.BoshTask{State: boshclient.BoshTaskError, Result: "not good", Description: "it's a task", ID: taskID},
					ActualOperationType: broker.OperationTypeCreate,
					LogContains:         "not good",

					ExpectedLastOperation: brokerapi.LastOperation{
						State:       brokerapi.Failed,
						Description: "Instance provisioning failed: An internal error has occurred",
					},
					TaskIDRemoved: true,
				}))

			Describe("last operation is Successful",
				testLastOperation(testCase{
					ActualBoshTask:      boshclient.BoshTask{State: boshclient.BoshTaskDone, Description: "it's a task", ID: taskID},
					ActualOperationType: broker.OperationTypeCreate,

					ExpectedLastOperation: brokerapi.LastOperation{
						State:       brokerapi.Succeeded,
						Description: broker.CreateCompleted,
					},
					TaskIDRemoved: true,
				}))
		})

		Describe("while deleting ", func() {
			Describe("last operation is Processing",
				testLastOperation(testCase{
					ActualBoshTask:      boshclient.BoshTask{State: boshclient.BoshTaskProcessing, Description: "it's a task" + "-" + instanceID},
					ActualOperationType: broker.OperationTypeDelete,

					ExpectedLastOperation: brokerapi.LastOperation{
						State:       brokerapi.InProgress,
						Description: broker.DeleteInProgress,
					},
				}))
			Describe("last operation is Queued",
				testLastOperation(testCase{
					ActualBoshTask:      boshclient.BoshTask{State: boshclient.BoshTaskQueued, Description: "it's a task" + "-" + instanceID},
					ActualOperationType: broker.OperationTypeDelete,

					ExpectedLastOperation: brokerapi.LastOperation{
						State:       brokerapi.InProgress,
						Description: broker.DeleteInProgress,
					},
				}))

			Describe("last operation is Error",
				testLastOperation(testCase{
					ActualBoshTask:      boshclient.BoshTask{State: boshclient.BoshTaskError, Result: "not good", Description: "it's a task" + "-" + instanceID, ID: taskID},
					ActualOperationType: broker.OperationTypeDelete,
					LogContains:         "not good",

					ExpectedLastOperation: brokerapi.LastOperation{
						State:       brokerapi.Failed,
						Description: "Instance deletion failed: An internal error has occurred",
					},
					TaskIDRemoved: true,
				}))

			Describe("last operation is Successful",
				testLastOperation(testCase{
					ActualBoshTask:      boshclient.BoshTask{State: boshclient.BoshTaskDone, Description: "it's a task" + "-" + instanceID, ID: taskID},
					ActualOperationType: broker.OperationTypeDelete,

					ExpectedLastOperation: brokerapi.LastOperation{
						State:       brokerapi.Succeeded,
						Description: broker.DeleteCompleted,
					},
				}))
		})

		Describe("while updating", func() {
			Describe("last operation is Processing",
				testLastOperation(testCase{
					ActualBoshTask:      boshclient.BoshTask{State: boshclient.BoshTaskProcessing, Description: "it's a task", ID: taskID},
					ActualOperationType: broker.OperationTypeUpdate,

					ExpectedLastOperation: brokerapi.LastOperation{
						State:       brokerapi.InProgress,
						Description: broker.UpdateInProgress,
					},
				}))
			Describe("last operation is Queued",
				testLastOperation(testCase{
					ActualBoshTask:      boshclient.BoshTask{State: boshclient.BoshTaskQueued, Description: "it's a task", ID: taskID},
					ActualOperationType: broker.OperationTypeUpdate,

					ExpectedLastOperation: brokerapi.LastOperation{
						State:       brokerapi.InProgress,
						Description: broker.UpdateInProgress,
					},
				}))

			Describe("last operation is Error",
				testLastOperation(testCase{
					ActualBoshTask:      boshclient.BoshTask{State: boshclient.BoshTaskError, Result: "not good", Description: "it's a task", ID: taskID},
					ActualOperationType: broker.OperationTypeUpdate,
					LogContains:         "not good",

					ExpectedLastOperation: brokerapi.LastOperation{
						State:       brokerapi.Failed,
						Description: "Instance update failed: An internal error has occurred",
					},
					TaskIDRemoved: true,
				}))

			Describe("last operation is Successful",
				testLastOperation(testCase{
					ActualBoshTask:      boshclient.BoshTask{State: boshclient.BoshTaskDone, Description: "it's a task", ID: taskID},
					ActualOperationType: broker.OperationTypeUpdate,

					ExpectedLastOperation: brokerapi.LastOperation{
						State:       brokerapi.Succeeded,
						Description: broker.UpdateCompleted,
					},
					TaskIDRemoved: true,
				}))
		})

		Describe("while upgrading ", func() {
			Describe("last operation is Processing",
				testLastOperation(testCase{
					ActualBoshTask:      boshclient.BoshTask{State: boshclient.BoshTaskProcessing, Description: "it's a task", ID: taskID},
					ActualOperationType: broker.OperationTypeUpgrade,

					ExpectedLastOperation: brokerapi.LastOperation{
						State: brokerapi.InProgress,
					},
				}))
			Describe("last operation is Queued",
				testLastOperation(testCase{
					ActualBoshTask:      boshclient.BoshTask{State: boshclient.BoshTaskQueued, Description: "it's a task", ID: taskID},
					ActualOperationType: broker.OperationTypeUpgrade,

					ExpectedLastOperation: brokerapi.LastOperation{
						State: brokerapi.InProgress,
					},
				}))

			Describe("last operation is Error",
				testLastOperation(testCase{
					ActualBoshTask:      boshclient.BoshTask{State: boshclient.BoshTaskError, Result: "not good", Description: "it's a task", ID: taskID},
					ActualOperationType: broker.OperationTypeUpgrade,

					ExpectedLastOperation: brokerapi.LastOperation{
						State:       brokerapi.Failed,
						Description: fmt.Sprintf("Failed for bosh task: 199"),
					},
					TaskIDRemoved: true,
				}))

			Describe("last operation is Successful",
				testLastOperation(testCase{
					ActualBoshTask:      boshclient.BoshTask{State: boshclient.BoshTaskDone, Description: "it's a task", ID: taskID},
					ActualOperationType: broker.OperationTypeUpgrade,

					ExpectedLastOperation: brokerapi.LastOperation{
						State: brokerapi.Succeeded,
					},
					TaskIDRemoved: true,
				}))
		})
	})
})
