package broker_test

import (
	"errors"

	"github.com/pivotal-cf/on-demand-service-broker/broker"
	"github.com/pivotal-cf/on-demand-service-broker/mgmtapi"

	. "github.com/onsi/ginkgo"
	. "github.com/onsi/gomega"
)

var _ = Describe("Upgrade", func() {
	var (
		upgradeTaskID int
		instanceID    = "some-instance"

		context = ItRedeploysContext{
			instanceID:    instanceID,
			oldPlanID:     existingPlanID,
			newPlanID:     existingPlanID,
			operationType: broker.OperationTypeUpgrade,
			boshTaskID:    876,
		}
	)

	JustBeforeEach(func() {
		upgradeTaskID, context.redeployErr = b.Upgrade(context.instanceID)
	})

	ItRedeploys(&context)

	Context("when there is a previous deployment for the service instance", func() {
		BeforeEach(func() {
			serviceAdapter.GenerateManifestReturns([]byte("fancy-new-manifest"), nil)
			boshClient.GetManifestRawReturns([]byte("crusty-old-manifest"), nil)
			fakeDeployer.DeployReturns(context.boshTaskID, []byte("new-manifest-fetched-from-adapter"), nil)
			cfClient.GetInstanceStateReturns(broker.InstanceState{PlanID: existingPlanID}, nil)
		})

		It("returns the task ID for the upgrade task", func() {
			Expect(upgradeTaskID).To(Equal(context.boshTaskID))
		})

		It("fetches correct instance", func() {
			Expect(cfClient.GetInstanceStateCallCount()).To(Equal(1))
			Expect(cfClient.GetInstanceStateArgsForCall(0)).To(Equal(instanceID))
		})

		Context("and there is an operation in progress on the instance", func() {
			BeforeEach(func() {
				cfClient.GetInstanceStateReturns(broker.InstanceState{PlanID: existingPlanID, OperationInProgress: true}, nil)
			})

			It("returns a deployment in progress error", func() {
				Expect(context.redeployErr).To(BeAssignableToTypeOf(mgmtapi.OperationInProgressError{}))
			})
		})
	})

	Context("when the service instance does not exist in Cloud Controller", func() {
		BeforeEach(func() {
			boshClient.GetManifestRawReturns([]byte("crusty-old-manifest"), nil)
			cfClient.GetInstanceStateReturns(broker.InstanceState{}, broker.NewNotFoundError())
		})

		It("returns a DeploymentNotFoundError", func() {
			Expect(context.redeployErr).To(BeAssignableToTypeOf(mgmtapi.DeploymentNotFoundError{}))
		})
	})

	Context("when the plan details cannot be retrieved from Cloud Controller", func() {
		BeforeEach(func() {
			boshClient.GetManifestRawReturns([]byte("crusty-old-manifest"), nil)
			cfClient.GetInstanceStateReturns(broker.InstanceState{}, errors.New("no instance for you"))
		})

		It("returns a generic error message to the cli user", func() {
			Expect(context.redeployErr).To(MatchError(ContainSubstring("no instance for you")))
		})
	})
})
