package cloud_foundry_client_test

import (
	"errors"
	"io/ioutil"
	"log"
	"os"
	"path"

	. "github.com/onsi/ginkgo"
	. "github.com/onsi/gomega"
	"github.com/pivotal-cf/on-demand-service-broker/authorizationheader/fake_auth_header_builder"
	"github.com/pivotal-cf/on-demand-service-broker/broker"
	"github.com/pivotal-cf/on-demand-service-broker/cloud_foundry_client"
	. "github.com/pivotal-cf/on-demand-service-broker/mock_cloud_foundry"
	"github.com/pivotal-cf/on-demand-service-broker/mockhttp"
)

var _ = Describe("Client", func() {
	var server *mockhttp.Server
	var testLogger *log.Logger
	cfAuthorizationHeader := "auth-header"
	var authHeaderBuilder *fake_auth_header_builder.FakeAuthHeaderBuilder

	BeforeEach(func() {
		authHeaderBuilder = new(fake_auth_header_builder.FakeAuthHeaderBuilder)
		authHeaderBuilder.BuildReturns(cfAuthorizationHeader, nil)
		server = NewMockCloudFoundry()
		testLogger = log.New(GinkgoWriter, "my-app", log.LstdFlags)
	})

	AfterEach(func() {
		server.VerifyMocks()
	})

	Describe("CountInstancesOfServiceOffering", func() {
		It("fetches instance counts per plan", func() {
			server.VerifyAndMock(
				ListServices().WithAuthorizationHeader(cfAuthorizationHeader).RespondsWith(fixture("list_services_response.json")),
				ListServicePlans("06df08f9-5a58-4d33-8097-32d0baf3ce1e").WithAuthorizationHeader(cfAuthorizationHeader).RespondsWith(fixture("list_service_plans_response.json")),
				ListServiceInstances("ff717e7c-afd5-4d0a-bafe-16c7eff546ec").WithAuthorizationHeader(cfAuthorizationHeader).RespondsWith(fixture("list_service_instances_for_plan_1_response.json")),
				ListServiceInstances("2777ad05-8114-4169-8188-2ef5f39e0c6b").WithAuthorizationHeader(cfAuthorizationHeader).RespondsWith(fixture("list_service_instances_for_plan_2_response.json")),
			)

			client := cloud_foundry_client.New(server.URL, testLogger, authHeaderBuilder, nil, true)

			Expect(client.CountInstancesOfServiceOffering("D94A086D-203D-4966-A6F1-60A9E2300F72")).To(Equal(map[string]int{
				"11789210-D743-4C65-9D38-C80B29F4D9C8": 1,
				"22789210-D743-4C65-9D38-C80B29F4D9C8": 2,
			}))
		})

		It("finds no instances when the service is not registered with cf", func() {
			server.VerifyAndMock(
				ListServices().WithAuthorizationHeader(cfAuthorizationHeader).RespondsWith(fixture("list_services_empty_response.json")),
			)

			client := cloud_foundry_client.New(server.URL, testLogger, authHeaderBuilder, nil, true)

			Expect(client.CountInstancesOfServiceOffering("D94A086D-203D-4966-A6F1-60A9E2300F72")).To(Equal(map[string]int{}))
		})

		It("fails if getting a new token fails", func() {
			server.VerifyAndMock(
				ListServices().WithAuthorizationHeader(cfAuthorizationHeader).RespondsWith(fixture("list_services_response.json")),
				ListServicePlans("06df08f9-5a58-4d33-8097-32d0baf3ce1e").WithAuthorizationHeader(cfAuthorizationHeader).RespondsWith(fixture("list_service_plans_response.json")),
				ListServiceInstances("ff717e7c-afd5-4d0a-bafe-16c7eff546ec").WithAuthorizationHeader(cfAuthorizationHeader).RespondsWithUnauthorized(`{"code": 1000,"description": "Invalid Auth Token","error_code": "CF-InvalidAuthToken"}`),
			)

			client := cloud_foundry_client.New(server.URL, testLogger, authHeaderBuilder, nil, true)

			_, err := client.CountInstancesOfServiceOffering("D94A086D-203D-4966-A6F1-60A9E2300F72")
			Expect(err).To(HaveOccurred())
			Expect(err).To(MatchError(ContainSubstring("Invalid Auth Token")))
		})

		It("reuses tokens", func() {
			server.VerifyAndMock(
				ListServices().WithAuthorizationHeader(cfAuthorizationHeader).RespondsWith(fixture("list_services_response.json")),
				ListServicePlans("06df08f9-5a58-4d33-8097-32d0baf3ce1e").WithAuthorizationHeader(cfAuthorizationHeader).RespondsWith(fixture("list_service_plans_response.json")),
				ListServiceInstances("ff717e7c-afd5-4d0a-bafe-16c7eff546ec").WithAuthorizationHeader(cfAuthorizationHeader).RespondsWith(fixture("list_service_instances_for_plan_1_response.json")),
				ListServiceInstances("2777ad05-8114-4169-8188-2ef5f39e0c6b").WithAuthorizationHeader(cfAuthorizationHeader).RespondsWith(fixture("list_service_instances_for_plan_2_response.json")),
			)

			client := cloud_foundry_client.New(server.URL, testLogger, authHeaderBuilder, nil, true)

			Expect(client.CountInstancesOfServiceOffering("D94A086D-203D-4966-A6F1-60A9E2300F72")).To(Equal(map[string]int{
				"11789210-D743-4C65-9D38-C80B29F4D9C8": 1,
				"22789210-D743-4C65-9D38-C80B29F4D9C8": 2,
			}))

			server.VerifyAndMock(
				ListServices().WithAuthorizationHeader(cfAuthorizationHeader).RespondsWith(fixture("list_services_response.json")),
				ListServicePlans("06df08f9-5a58-4d33-8097-32d0baf3ce1e").WithAuthorizationHeader(cfAuthorizationHeader).RespondsWith(fixture("list_service_plans_response.json")),
				ListServiceInstances("ff717e7c-afd5-4d0a-bafe-16c7eff546ec").WithAuthorizationHeader(cfAuthorizationHeader).RespondsWith(fixture("list_service_instances_for_plan_1_response.json")),
				ListServiceInstances("2777ad05-8114-4169-8188-2ef5f39e0c6b").WithAuthorizationHeader(cfAuthorizationHeader).RespondsWith(fixture("list_service_instances_for_plan_2_response.json")),
			)

			Expect(client.CountInstancesOfServiceOffering("D94A086D-203D-4966-A6F1-60A9E2300F72")).To(Equal(map[string]int{
				"11789210-D743-4C65-9D38-C80B29F4D9C8": 1,
				"22789210-D743-4C65-9D38-C80B29F4D9C8": 2,
			}))
		})

		It("fetches instance counts per plan, across service pages", func() {
			server.VerifyAndMock(
				ListServices().WithAuthorizationHeader(cfAuthorizationHeader).RespondsWith(fixture("list_services_response_page_1.json")),
				ListServicesForPage(2).WithAuthorizationHeader(cfAuthorizationHeader).RespondsWith(fixture("list_services_response_page_2.json")),
				ListServicePlans("06df08f9-5a58-4d33-8097-32d0baf3ce1e").WithAuthorizationHeader(cfAuthorizationHeader).RespondsWith(fixture("list_service_plans_response.json")),
				ListServiceInstances("ff717e7c-afd5-4d0a-bafe-16c7eff546ec").WithAuthorizationHeader(cfAuthorizationHeader).RespondsWith(fixture("list_service_instances_for_plan_1_response.json")),
				ListServiceInstances("2777ad05-8114-4169-8188-2ef5f39e0c6b").WithAuthorizationHeader(cfAuthorizationHeader).RespondsWith(fixture("list_service_instances_for_plan_2_response.json")),
			)

			client := cloud_foundry_client.New(server.URL, testLogger, authHeaderBuilder, nil, true)

			Expect(client.CountInstancesOfServiceOffering("D94A086D-203D-4966-A6F1-60A9E2300F72")).To(Equal(map[string]int{
				"11789210-D743-4C65-9D38-C80B29F4D9C8": 1,
				"22789210-D743-4C65-9D38-C80B29F4D9C8": 2,
			}))
		})

		It("fetches instance counts per plan, across plan pages", func() {
			server.VerifyAndMock(
				ListServices().WithAuthorizationHeader(cfAuthorizationHeader).RespondsWith(fixture("list_services_response.json")),
				ListServicePlans("06df08f9-5a58-4d33-8097-32d0baf3ce1e").WithAuthorizationHeader(cfAuthorizationHeader).RespondsWith(fixture("list_service_plans_response_page_1.json")),
				ListServicePlansForPage("06df08f9-5a58-4d33-8097-32d0baf3ce1e", 2).WithAuthorizationHeader(cfAuthorizationHeader).RespondsWith(fixture("list_service_plans_response_page_2.json")),
				ListServiceInstances("ff717e7c-afd5-4d0a-bafe-16c7eff546ec").WithAuthorizationHeader(cfAuthorizationHeader).RespondsWith(fixture("list_service_instances_for_plan_1_response.json")),
				ListServiceInstances("2777ad05-8114-4169-8188-2ef5f39e0c6b").WithAuthorizationHeader(cfAuthorizationHeader).RespondsWith(fixture("list_service_instances_for_plan_2_response.json")),
			)

			client := cloud_foundry_client.New(server.URL, testLogger, authHeaderBuilder, nil, true)

			Expect(client.CountInstancesOfServiceOffering("D94A086D-203D-4966-A6F1-60A9E2300F72")).To(Equal(map[string]int{
				"11789210-D743-4C65-9D38-C80B29F4D9C8": 1,
				"22789210-D743-4C65-9D38-C80B29F4D9C8": 2,
			}))
		})

		It("fails, if fetching auth token fails", func() {
			authHeaderBuilder.BuildReturns("", errors.New("niet goed"))

			client := cloud_foundry_client.New(server.URL, testLogger, authHeaderBuilder, nil, true)

			_, err := client.GetInstancesOfServiceOffering("some-offering")
			Expect(err).To(MatchError(ContainSubstring("niet goed")))
		})

		It("fails if fetching services fails", func() {
			server.VerifyAndMock(
				ListServices().Fails("niet goed"),
			)

			client := cloud_foundry_client.New(server.URL, testLogger, authHeaderBuilder, nil, true)

			_, err := client.CountInstancesOfServiceOffering("D94A086D-203D-4966-A6F1-60A9E2300F72")
			Expect(err).To(MatchError(ContainSubstring("niet goed")))
		})

		It("fails if fetching services fails", func() {
			server.VerifyAndMock(
				ListServices().Fails("niet goed"),
			)

			client := cloud_foundry_client.New(server.URL, testLogger, authHeaderBuilder, nil, true)

			_, err := client.CountInstancesOfServiceOffering("D94A086D-203D-4966-A6F1-60A9E2300F72")
			Expect(err).To(MatchError(ContainSubstring("niet goed")))
		})

		It("fails if fetching service plans fails", func() {
			server.VerifyAndMock(
				ListServices().WithAuthorizationHeader(cfAuthorizationHeader).RespondsWith(fixture("list_services_response.json")),
				ListServicePlans("06df08f9-5a58-4d33-8097-32d0baf3ce1e").Fails("niet goed"),
			)

			client := cloud_foundry_client.New(server.URL, testLogger, authHeaderBuilder, nil, true)

			_, err := client.CountInstancesOfServiceOffering("D94A086D-203D-4966-A6F1-60A9E2300F72")
			Expect(err).To(MatchError(ContainSubstring("niet goed")))
		})

		It("fails if fetching service instances fails", func() {
			server.VerifyAndMock(
				ListServices().WithAuthorizationHeader(cfAuthorizationHeader).RespondsWith(fixture("list_services_response.json")),
				ListServicePlans("06df08f9-5a58-4d33-8097-32d0baf3ce1e").WithAuthorizationHeader(cfAuthorizationHeader).RespondsWith(fixture("list_service_plans_response.json")),
				ListServiceInstances("ff717e7c-afd5-4d0a-bafe-16c7eff546ec").Fails("niet goed"),
			)

			client := cloud_foundry_client.New(server.URL, testLogger, authHeaderBuilder, nil, true)

			_, err := client.CountInstancesOfServiceOffering("D94A086D-203D-4966-A6F1-60A9E2300F72")
			Expect(err).To(MatchError(ContainSubstring("niet goed")))
		})

	})
	Describe("CountInstancesOfPlan", func() {
		It("fetches instance counts for the plan", func() {
			server.VerifyAndMock(
				ListServices().WithAuthorizationHeader(cfAuthorizationHeader).RespondsWith(fixture("list_services_response.json")),
				ListServicePlans("06df08f9-5a58-4d33-8097-32d0baf3ce1e").WithAuthorizationHeader(cfAuthorizationHeader).RespondsWith(fixture("list_service_plans_response.json")),
				ListServiceInstances("2777ad05-8114-4169-8188-2ef5f39e0c6b").WithAuthorizationHeader(cfAuthorizationHeader).RespondsWith(fixture("list_service_instances_for_plan_2_response.json")),
			)

			client := cloud_foundry_client.New(server.URL, testLogger, authHeaderBuilder, nil, true)

			Expect(client.CountInstancesOfPlan("D94A086D-203D-4966-A6F1-60A9E2300F72", "22789210-D743-4C65-9D38-C80B29F4D9C8")).To(Equal(2))
		})

		It("fail if service instance not found", func() {
			server.VerifyAndMock(
				ListServices().WithAuthorizationHeader(cfAuthorizationHeader).RespondsWith(fixture("list_services_response.json")),
				ListServicePlans("06df08f9-5a58-4d33-8097-32d0baf3ce1e").WithAuthorizationHeader(cfAuthorizationHeader).RespondsWith(fixture("list_service_plans_response.json")),
			)

			client := cloud_foundry_client.New(server.URL, testLogger, authHeaderBuilder, nil, true)

			count, err := client.CountInstancesOfPlan("D94A086D-203D-4966-A6F1-60A9E2300F72", "does-not-exist")
			Expect(err).To(MatchError(ContainSubstring("service plan does-not-exist not found")))
			Expect(count).To(BeZero())
		})

		It("fails when it can't retrieve services", func() {
			server.VerifyAndMock(
				ListServices().WithAuthorizationHeader(cfAuthorizationHeader).Fails("no services for you"),
			)

			client := cloud_foundry_client.New(server.URL, testLogger, authHeaderBuilder, nil, true)

			count, err := client.CountInstancesOfPlan("D94A086D-203D-4966-A6F1-60A9E2300F72", "22789210-D743-4C65-9D38-C80B29F4D9C8")
			Expect(err).To(MatchError(ContainSubstring("no services for you")))
			Expect(count).To(BeZero())
		})

		It("fails when it can't retrieve service plans", func() {
			server.VerifyAndMock(
				ListServices().WithAuthorizationHeader(cfAuthorizationHeader).RespondsWith(fixture("list_services_response.json")),
				ListServicePlans("06df08f9-5a58-4d33-8097-32d0baf3ce1e").Fails("no service plans for you"),
			)

			client := cloud_foundry_client.New(server.URL, testLogger, authHeaderBuilder, nil, true)

			count, err := client.CountInstancesOfPlan("D94A086D-203D-4966-A6F1-60A9E2300F72", "22789210-D743-4C65-9D38-C80B29F4D9C8")
			Expect(err).To(MatchError(ContainSubstring("no service plans for you")))
			Expect(count).To(BeZero())
		})

		It("fails when it can't retrieve service instnaces for the plan", func() {
			server.VerifyAndMock(
				ListServices().WithAuthorizationHeader(cfAuthorizationHeader).RespondsWith(fixture("list_services_response.json")),
				ListServicePlans("06df08f9-5a58-4d33-8097-32d0baf3ce1e").WithAuthorizationHeader(cfAuthorizationHeader).RespondsWith(fixture("list_service_plans_response.json")),
				ListServiceInstances("2777ad05-8114-4169-8188-2ef5f39e0c6b").Fails("no instances for you"),
			)

			client := cloud_foundry_client.New(server.URL, testLogger, authHeaderBuilder, nil, true)

			count, err := client.CountInstancesOfPlan("D94A086D-203D-4966-A6F1-60A9E2300F72", "22789210-D743-4C65-9D38-C80B29F4D9C8")
			Expect(err).To(MatchError(ContainSubstring("no instances for you")))
			Expect(count).To(BeZero())
		})
	})

	Describe("GetInstanceState", func() {
		It("fetches the state of an instance", func() {
			server.VerifyAndMock(
				GetServiceInstance("783f8645-1ded-4161-b457-73f59423f9eb").RespondsWith(fixture("get_service_instance_response.json")),
				GetServicePlan("ff717e7c-afd5-4d0a-bafe-16c7eff546ec").RespondsWith(fixture("get_service_plan_response.json")),
			)

			client := cloud_foundry_client.New(server.URL, testLogger, authHeaderBuilder, nil, true)

			state, err := client.GetInstanceState("783f8645-1ded-4161-b457-73f59423f9eb")
			Expect(state.PlanID).To(Equal("11789210-D743-4C65-9D38-C80B29F4D9C8"))
			Expect(state.OperationInProgress).To(BeFalse())
			Expect(err).NotTo(HaveOccurred())
		})

		It("returns that an operation is in progress when an operation is in progress", func() {
			server.VerifyAndMock(
				GetServiceInstance("783f8645-1ded-4161-b457-73f59423f9eb").RespondsWith(fixture("get_service_instance_operation_in_progress_response.json")),
				GetServicePlan("ff717e7c-afd5-4d0a-bafe-16c7eff546ec").RespondsWith(fixture("get_service_plan_response.json")),
			)

			client := cloud_foundry_client.New(server.URL, testLogger, authHeaderBuilder, nil, true)

			state, err := client.GetInstanceState("783f8645-1ded-4161-b457-73f59423f9eb")
			Expect(state.PlanID).To(Equal("11789210-D743-4C65-9D38-C80B29F4D9C8"))
			Expect(state.OperationInProgress).To(BeTrue())
			Expect(err).NotTo(HaveOccurred())
		})

		Context("when the service instance cannot be retrieved", func() {
			It("returns an error", func() {
				server.VerifyAndMock(
					GetServiceInstance("783f8645-1ded-4161-b457-73f59423f9eb").Fails("er ma gerd"),
				)

				client := cloud_foundry_client.New(server.URL, testLogger, authHeaderBuilder, nil, true)

				_, err := client.GetInstanceState("783f8645-1ded-4161-b457-73f59423f9eb")
				Expect(err).To(MatchError(ContainSubstring("er ma gerd")))
				Expect(broker.IsNotFound(err)).To(BeFalse())
			})
		})

		Context("when the service instance does not exist", func() {
			It("returns a not found error", func() {
				server.VerifyAndMock(
					GetServiceInstance("783f8645-1ded-4161-b457-73f59423f9eb").NotFound(),
				)

				client := cloud_foundry_client.New(server.URL, testLogger, authHeaderBuilder, nil, true)

				_, err := client.GetInstanceState("783f8645-1ded-4161-b457-73f59423f9eb")
				Expect(broker.IsNotFound(err)).To(BeTrue())
			})
		})

		Context("when the service plan cannot be retrieved", func() {
			It("returns an error", func() {
				server.VerifyAndMock(
					GetServiceInstance("783f8645-1ded-4161-b457-73f59423f9eb").RespondsWith(fixture("get_service_instance_response.json")),
					GetServicePlan("ff717e7c-afd5-4d0a-bafe-16c7eff546ec").Fails("er ma gerd"),
				)

				client := cloud_foundry_client.New(server.URL, testLogger, authHeaderBuilder, nil, true)

				_, err := client.GetInstanceState("783f8645-1ded-4161-b457-73f59423f9eb")
				Expect(err).To(MatchError(ContainSubstring("er ma gerd")))
				Expect(broker.IsNotFound(err)).To(BeFalse())
			})
		})

		Context("when the service plan does not exist", func() {
			It("returns a not found error", func() {
				server.VerifyAndMock(
					GetServiceInstance("783f8645-1ded-4161-b457-73f59423f9eb").RespondsWith(fixture("get_service_instance_response.json")),
					GetServicePlan("ff717e7c-afd5-4d0a-bafe-16c7eff546ec").NotFound(),
				)

				client := cloud_foundry_client.New(server.URL, testLogger, authHeaderBuilder, nil, true)

				_, err := client.GetInstanceState("783f8645-1ded-4161-b457-73f59423f9eb")
				Expect(broker.IsNotFound(err)).To(BeTrue())
			})
		})
	})

	Describe("GetInstancesOfServiceOffering", func() {
		It("returns a list of instance IDs", func() {
			offeringID := "8F3E8998-5FD0-4F32-924A-5478DC390A5F"

			server.VerifyAndMock(

				ListServices().WithAuthorizationHeader(cfAuthorizationHeader).RespondsWith(fixture("list_services_response.json")),
				ListServicePlans("34c08156-5b5d-4cc1-9af1-29cda9ec056f").WithAuthorizationHeader(cfAuthorizationHeader).RespondsWith(fixture("list_service_plans_response.json")),
				ListServiceInstances("ff717e7c-afd5-4d0a-bafe-16c7eff546ec").WithAuthorizationHeader(cfAuthorizationHeader).RespondsWith(fixture("list_service_instances_for_plan_1_response.json")),
				ListServiceInstances("2777ad05-8114-4169-8188-2ef5f39e0c6b").WithAuthorizationHeader(cfAuthorizationHeader).RespondsWith(fixture("list_service_instances_for_plan_2_response.json")))

			client := cloud_foundry_client.New(server.URL, testLogger, authHeaderBuilder, nil, true)

			instances, err := client.GetInstancesOfServiceOffering(offeringID)
			Expect(err).NotTo(HaveOccurred())
			Expect(instances).To(ConsistOf("520f8566-b727-4c67-8be8-d9285645e936", "f897f40d-0b2d-474a-a5c9-98426a2cb4b8", "2f759033-04a4-426b-bccd-01722036c152"))
		})

		Context("when the list of services cannot be retrieved", func() {
			It("returns an error", func() {
				offeringID := "8F3E8998-5FD0-4F32-924A-5478DC390A5F"

				server.VerifyAndMock(

					ListServices().WithAuthorizationHeader(cfAuthorizationHeader).Fails("oops"),
				)

				client := cloud_foundry_client.New(server.URL, testLogger, authHeaderBuilder, nil, true)

				_, err := client.GetInstancesOfServiceOffering(offeringID)
				Expect(err).To(MatchError(ContainSubstring("oops")))
			})
		})

		Context("when the list of plans for the service cannot be retrieved", func() {
			It("returns an error", func() {
				offeringID := "8F3E8998-5FD0-4F32-924A-5478DC390A5F"

				server.VerifyAndMock(

					ListServices().WithAuthorizationHeader(cfAuthorizationHeader).RespondsWith(fixture("list_services_response.json")),
					ListServicePlans("34c08156-5b5d-4cc1-9af1-29cda9ec056f").WithAuthorizationHeader(cfAuthorizationHeader).Fails("oops"),
				)

				client := cloud_foundry_client.New(server.URL, testLogger, authHeaderBuilder, nil, true)

				_, err := client.GetInstancesOfServiceOffering(offeringID)
				Expect(err).To(MatchError(ContainSubstring("oops")))
			})
		})

		Context("when the list of instances for a plan cannot be retrieved", func() {
			It("returns an error", func() {
				offeringID := "8F3E8998-5FD0-4F32-924A-5478DC390A5F"

				server.VerifyAndMock(

					ListServices().WithAuthorizationHeader(cfAuthorizationHeader).RespondsWith(fixture("list_services_response.json")),
					ListServicePlans("34c08156-5b5d-4cc1-9af1-29cda9ec056f").WithAuthorizationHeader(cfAuthorizationHeader).RespondsWith(fixture("list_service_plans_response.json")),
					ListServiceInstances("ff717e7c-afd5-4d0a-bafe-16c7eff546ec").WithAuthorizationHeader(cfAuthorizationHeader).Fails("oops"),
				)

				client := cloud_foundry_client.New(server.URL, testLogger, authHeaderBuilder, nil, true)

				_, err := client.GetInstancesOfServiceOffering(offeringID)
				Expect(err).To(MatchError(ContainSubstring("oops")))
			})
		})

	})
})

func fixture(filename string) string {
	file, err := os.Open(path.Join("fixtures", filename))
	Expect(err).NotTo(HaveOccurred())
	content, err := ioutil.ReadAll(file)
	Expect(err).NotTo(HaveOccurred())
	return string(content)
}
