package external

import (
	"encoding/json"
	"errors"
	"fmt"
	"log"

	"gopkg.in/yaml.v2"

	"github.com/pivotal-cf/on-demand-service-broker-sdk/bosh"
	"github.com/pivotal-cf/on-demand-service-broker-sdk/serviceadapter"
	"github.com/pivotal-cf/on-demand-service-broker/broker"
)

const ExitCodeOperationNotImplemented = 10
const ExitCodeBindingAppGuidNotSpecified = 42
const ExitCodeBindingAlreadyExists = 49

type Adapter struct {
	ExternalBinPath string
	CommandRunner   CommandRunner
	Logger          *log.Logger
}

func (a *Adapter) GenerateManifest(serviceDeployment serviceadapter.ServiceDeployment, plan serviceadapter.Plan, requestParams map[string]interface{}, previousManifest []byte, previousPlan *serviceadapter.Plan) ([]byte, error) {
	errs := func(err error) ([]byte, error) {
		return nil, err
	}

	serialisedServiceDeployment, err := json.Marshal(serviceDeployment)
	if err != nil {
		return errs(err)
	}

	plan.Properties = SanitiseForJSON(plan.Properties)
	serialisedPlan, err := json.Marshal(plan)
	if err != nil {
		return errs(err)
	}

	serialisedRequestParams, err := json.Marshal(requestParams)
	if err != nil {
		return errs(err)
	}
	if previousPlan != nil {
		previousPlan.Properties = SanitiseForJSON(previousPlan.Properties)
	}

	serialisedPreviousPlan, err := json.Marshal(previousPlan)
	if err != nil {
		return errs(err)
	}

	stdout, stderr, exitCode, err := a.CommandRunner.Run(a.ExternalBinPath, "generate-manifest", string(serialisedServiceDeployment),
		string(serialisedPlan), string(serialisedRequestParams), string(previousManifest), string(serialisedPreviousPlan))

	if err != nil {
		return errs(a.failWithAdapterError(err, stdout, stderr))
	}

	switch *exitCode {
	case 0:
		break
	case ExitCodeOperationNotImplemented:
		a.Logger.Println("service adapter has not implemented generate-manifest command")
		return errs(a.failWithExitCodeError(err, *exitCode, stdout, stderr))
	default:
		return errs(a.failWithExitCodeError(err, *exitCode, stdout, stderr))
	}

	a.Logger.Printf("service adapter ran generate-manifest successfully, stderr logs: %s", string(stderr))

	var generatedManifest struct {
		Name string `yaml:"name"`
	}

	if err := yaml.Unmarshal(stdout, &generatedManifest); err != nil {
		return errs(a.failWithInvalidYamlError(stderr))
	}

	if generatedManifest.Name != serviceDeployment.DeploymentName {
		a.Logger.Printf("service adapter generated a manifest with an incorrect deployment name: %s", generatedManifest.Name)
		return errs(a.failWithIncorrectDeploymentNameError(serviceDeployment.DeploymentName, generatedManifest.Name, stderr))
	}

	return stdout, nil
}

func (a *Adapter) failWithAdapterError(err error, stdout []byte, stderr []byte) error {
	return fmt.Errorf("an error occurred running external service adapter at %s: '%s'. Stdout: '%s', stderr: '%s'", a.ExternalBinPath, err, string(stdout), string(stderr))
}
func (a *Adapter) failWithIncorrectDeploymentNameError(expectedName, actualName string, stderr []byte) error {
	return fmt.Errorf("external service adapter generated manifest with an incorrect deployment name at %s. expected name: '%s', returned name: '%s', stderr: '%s'", a.ExternalBinPath, expectedName, actualName, string(stderr))
}
func (a *Adapter) failWithInvalidYamlError(stderr []byte) error {
	return fmt.Errorf("external service adapter generated manifest that is not valid YAML at %s. stderr: '%s'", a.ExternalBinPath, string(stderr))
}

func (a *Adapter) failWithExitCodeError(err error, exitCode int, stdout []byte, stderr []byte) error {
	messageForOperator := fmt.Sprintf("external service adapter exited with %d at %s: Stdout: '%s', stderr: '%s'", exitCode, a.ExternalBinPath, string(stdout), string(stderr))
	if string(stdout) == "" {
		return errors.New(messageForOperator)
	}
	return broker.NewAdapterCommandError(string(stdout), messageForOperator)
}

func SanitiseForJSON(properties serviceadapter.Properties) serviceadapter.Properties {
	propertiesToReturn := serviceadapter.Properties{}

	for k, v := range properties {
		propertiesToReturn[k] = sanitiseValueForJSON(v)
	}

	return propertiesToReturn
}

// structures parsed from YAML contain map[interface{}]interface{}, but JSON marshalling requires map[string]interface{}, so we have to walk the structure and convert any instances of the former to the latter
func sanitiseValueForJSON(obj interface{}) interface{} {
	switch castObj := obj.(type) {

	case map[interface{}]interface{}:
		sanitisedMap := map[string]interface{}{}
		for key, value := range castObj {
			sanitisedMap[key.(string)] = sanitiseValueForJSON(value)
		}
		return sanitisedMap

	case []interface{}:
		var sanitisedSlice []interface{}
		for _, value := range castObj {
			sanitisedSlice = append(sanitisedSlice, sanitiseValueForJSON(value))
		}
		return sanitisedSlice

	default:
		return obj

	}
}

func (a *Adapter) CreateBinding(bindingID string, deploymentTopology bosh.BoshVMs, manifest []byte, requestParams map[string]interface{}) (serviceadapter.Binding, error) {
	errs := func(err error) (serviceadapter.Binding, error) {
		return serviceadapter.Binding{}, err
	}

	serialisedBoshVMs, err := json.Marshal(deploymentTopology)
	if err != nil {
		return errs(err)
	}
	serialisedRequestParams, err := json.Marshal(requestParams)
	if err != nil {
		return errs(err)
	}

	stdout, stderr, exitcode, err := a.CommandRunner.Run(a.ExternalBinPath, "create-binding", bindingID, string(serialisedBoshVMs), string(manifest), string(serialisedRequestParams))
	if err != nil {
		return errs(a.failWithAdapterError(err, stdout, stderr))
	}

	exitError := a.failWithExitCodeError(err, *exitcode, stdout, stderr)
	switch *exitcode {
	case 0:
		break
	case ExitCodeOperationNotImplemented:
		a.Logger.Println("service adapter has not implemented create-binding command")
		return errs(exitError)
	case ExitCodeBindingAppGuidNotSpecified:
		a.Logger.Println(exitError.Error())
		return errs(serviceadapter.NewAppGuidNotProvidedError())
	case ExitCodeBindingAlreadyExists:
		a.Logger.Println(exitError.Error())
		return errs(serviceadapter.NewBindingAlreadyExistsError())
	default:
		return errs(exitError)
	}

	a.Logger.Printf("service adapter ran create-binding successfully, stderr logs: %s", string(stderr))

	var adapterBinding serviceadapter.Binding
	if err := json.Unmarshal(stdout, &adapterBinding); err != nil {
		return errs(err)
	}

	return adapterBinding, nil
}

func (a *Adapter) DeleteBinding(bindingID string, deploymentTopology bosh.BoshVMs, manifest []byte, requestParams map[string]interface{}) error {
	serialisedBoshVMs, err := json.Marshal(deploymentTopology)
	if err != nil {
		return err
	}

	serialisedRequestParams, err := json.Marshal(requestParams)
	if err != nil {
		return err
	}

	stdout, stderr, exitCode, err := a.CommandRunner.Run(a.ExternalBinPath, "delete-binding", bindingID, string(serialisedBoshVMs), string(manifest), string(serialisedRequestParams))
	if err != nil {
		return a.failWithAdapterError(err, stdout, stderr)
	}

	switch *exitCode {
	case 0:
		break
	case ExitCodeOperationNotImplemented:
		a.Logger.Println("service adapter has not implemented delete-binding command")
		return a.failWithExitCodeError(err, *exitCode, stdout, stderr)
	default:
		return a.failWithExitCodeError(err, *exitCode, stdout, stderr)
	}

	a.Logger.Printf("service adapter ran delete-binding successfully, stderr logs: %s", string(stderr))

	return nil
}

func (a *Adapter) GenerateDashboardUrl(instanceID string, plan serviceadapter.Plan, manifest []byte) (string, error) {
	plan.Properties = SanitiseForJSON(plan.Properties)
	planJSON, err := json.Marshal(plan)
	if err != nil {
		return "", err
	}

	stdout, stderr, exitCode, err := a.CommandRunner.Run(a.ExternalBinPath, "dashboard-url", instanceID, string(planJSON), string(manifest))

	if err != nil {
		return "", a.failWithAdapterError(err, stdout, stderr)
	}

	switch *exitCode {
	case 0:
		break
	case ExitCodeOperationNotImplemented:
		a.Logger.Println("service adapter has not implemented dashboard-url command")
		return "", broker.NewServiceAdapterCommandNotImplementedError(a.failWithExitCodeError(err, *exitCode, stdout, stderr))
	default:
		return "", a.failWithExitCodeError(err, *exitCode, stdout, stderr)
	}

	dashboardURL := serviceadapter.DashboardUrl{}

	return string(dashboardURL.DashboardUrl), json.Unmarshal(stdout, &dashboardURL)
}
