package mgmtapi

import (
	"encoding/json"
	"io"
	"log"
	"net/http"

	"github.com/gorilla/mux"
	"github.com/pivotal-cf/brokerapi"
)

type api struct {
	manageableBroker ManageableBroker
	logger           *log.Logger
}

type DeploymentNotFoundError struct {
	error
}

func NewDeploymentNotFoundError(e error) error {
	return DeploymentNotFoundError{e}
}

type OperationInProgressError struct {
	error
}

func NewOperationInProgressError(e error) error {
	return OperationInProgressError{e}
}

//go:generate counterfeiter -o fake_manageable_broker/fake_manageable_broker.go . ManageableBroker
type ManageableBroker interface {
	Instances() ([]Instance, error)
	Upgrade(instanceID string) (int, error)
}

type Instance struct {
	InstanceID string `json:"instance_id"`
}

type UpgradeResponse struct {
	TaskID int `json:"task_id"`
}

func AttachRoutes(r *mux.Router, manageableBroker ManageableBroker, logger *log.Logger) {
	a := &api{manageableBroker: manageableBroker, logger: logger}
	r.HandleFunc("/mgmt/service_instances", a.listAllInstances).Methods("GET")
	r.HandleFunc("/mgmt/service_instances/{instance_id}", a.upgradeInstance).Methods("PATCH")
}

func (a *api) listAllInstances(w http.ResponseWriter, r *http.Request) {
	instances, err := a.manageableBroker.Instances()
	if err != nil {
		a.logger.Printf("error occurred querying instances: %s", err)
		w.WriteHeader(http.StatusInternalServerError)
		return
	}
	a.writeJson(w, instances)
}

func (a *api) upgradeInstance(w http.ResponseWriter, r *http.Request) {
	vars := mux.Vars(r)
	instanceID := vars["instance_id"]
	boshTaskID, err := a.manageableBroker.Upgrade(instanceID)
	switch err.(type) {
	case nil:
		w.WriteHeader(http.StatusAccepted)
		a.writeJson(w, UpgradeResponse{TaskID: boshTaskID})
	case DeploymentNotFoundError:
		w.WriteHeader(http.StatusGone)
	case OperationInProgressError:
		w.WriteHeader(http.StatusConflict)
	case error:
		a.logger.Printf("error occurred upgrading instance %s: %s", instanceID, err)
		w.WriteHeader(http.StatusInternalServerError)
		a.writeJson(w, brokerapi.ErrorResponse{Description: err.Error()})
	}
}

func (a *api) writeJson(w io.Writer, obj interface{}) {
	if err := json.NewEncoder(w).Encode(obj); err != nil {
		a.logger.Printf("error occurred encoding json: %s", err)
	}
}
